/*
 * frost - C++ multi methods as if they were a native feature
 * Copyright (C) 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */

#ifndef _frost_as_hh
#define _frost_as_hh


// includes

#include <iostream>

using namespace std;


/** prints assembly code that defines the head of a frost table
	\param out output stream
	\param name table name
	\param sz number of table entries */
void
DefineFTableHead( ostream& out, const string& name, size_t sz );

/** prints assembly code that defines a frost table
	\param out output stream
	\param name table name */
void
DefineFTableDef( ostream& out, const string& name );

/** prints assembly code that defines a typeinfo string
	\param out output stream
	\param name virtual table name */
void
PrintTIString( ostream& out, const string& name );

/** prints assembly code that defines a virtual table
	\param out output stream
	\param name table name
	\param sz number of table entries */
void
DefineVTable( ostream& out, const string& name, size_t sz );

/** prints assembly code of a table entry
	\param out output stream
	\param v value */
void
PrintTableEntry( ostream& out, const string& v );

/** prints assembly code which enters the .text segment
	\param out output stream */
void
EnterTextSegment( ostream& out );

/** prints assembly code that defines a function
	\param out output stream
	\param name function name
	\param g 'global function' flag */
void
DefineFunction( ostream& out, const string& name, bool g );

/** prints assembly code that reads the 'this' pointer of an object from the
	stack, adjusts the pointer and stores it on the stack and (optional) the
	working register.
	\param out output stream
	\param so offset on the stack
	\param dt 'this' pointer delta
	\param wr 'store in working register' flag */
void
AdjustThis( ostream& out, int so, int dt, bool wr );

/** prints assembly code that adds a value to the stack pointer, dereferences
	the address, and stores the result in the working register
	\param out output stream
	\param dt delta */
void
StackDeref( ostream& out, int dt );

/** prints assembly code that adds a value to an address stored in the working
	register, dereferences the address, and stores the result in the working
	register
	\param out output stream
	\param dt delta */
void
WRegDeref( ostream& out, int dt );

/** prints assembly code that adds a value to an address stored in the working
	register, dereferences the address, and jumps to the resulting address
	\param out output stream
	\param dt delta */
void
WRegJump( ostream& out, int dt );

/** prints assembly code that jumps performs a jump to a symbol
	\param out output stream
	\param tgt target symbol */
void
Jump( ostream& out, const string& tgt );

/** prints code for a global string constant
	\param out output stream
	\param sym symbol name
	\param str string constant */
void
StrConstant( ostream& out, const char* sym, const string& str );

/** Prints code for a global chunk of binary data
	\param out output stream
	\param sym symbol name
	\param str string constant */
void
BinConstant( ostream& out, const char* sym, const string& str );

/** prints the 'noreturn' dummy
	\param out output stream */
void
EmitNoreturn( ostream& out );

#endif // _frost_as_hh
