/*
 * Frost - use C++ multi methods as if they were a native feature
 * Copyright (C) 2000, 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */

#ifndef _frost_class_hh
#define _frost_class_hh


// includes

#include <map>
#include <set>
#include <string>
#include <vector>

using namespace std;


class Table;
class TableGrp;
class VTable;


/** class class ;-)
	\author Julian v. Bock */
class Class
{	
public:
	typedef set< Class* > class_set;
	typedef set< Table* > table_set;

	/** \param n resolved class name
		\param sz size
		\param b 'builtin type' flag */
	Class( const string& n, size_t sz, bool b );

	/** \return class name */
	const string&
	Name() const { return name; }

	/** returns the encoded class name for use in fwva dispatcher names.
		note that the encoding scheme is not the same the compiler uses.
		\return encoded class name */
	string
	EName() const;

	/** \return class size */
	size_t
	Size() const { return size; }

	/** \return 'builtin type' flag */
	bool
	BIType() const { return bit; }

	/** \param sz class size */
	void
	SetSize( size_t sz );

	/** destroys the object */
	void
	Destroy() { delete this; }

	/** checks whether this class is the topmost element in a hierarchy
		\return 'topmost class' flag */
	bool
	Topmost() const { return bases.empty(); }
	
	/** calculates how good a given class matches this one.
		\param h class
		\param m this parameter is set to the distance between the types
		if they are in one hierarchy
		\return false if the classes are not in one hierarchy */
	bool
	Match( Class& h, int& m );

	/** returns the offset of a virtual table
		\param cl class
		\return table offset */
	size_t
	TableOffset( Class& cl ) const;

	/** sets a virtual table offset
		\param cl class
		\param o offset
		\param d distance to base class */
	void
	SetTableOffset( Class& cl, size_t o, int d );

	/** registers a virtual base class.
		\param cl class */
	void
	AddVBase( Class& cl );

	/** \return virtual table */
	const VTable&
	Vtbl() const;

	/** sets the virtual table
		\param vt virtual table */
	void
	SetVtbl( VTable& vt );

	/** creates and returns a new frost table
		\param n name
		\param tg table group
		\param fvo initial frost vtable offset
		\return new frost table */
	Table&
	AddFTable( const string& n, TableGrp& tg, size_t fvo );

	/** \return main frost table */
	Table&
	MainFTable();

	/** fills a set with the main table and all derived tables
		\param ts table set */
	void
	DTableSet( table_set& ts );

	/** mark the class as tainted. tainted classes must not be involved in
		fwvas. */
	void
	Taint();

	/** checks whether the class is tainted.
		\return 'tainted' flag */
	bool
	Tainted() const { return tainted; }

	/** output stream operator
		\param out output stream
		\param cl class */
	friend ostream&
	operator << ( ostream& out, const Class& cl );

private:
	/** virtual table / base info class
		\author Julian v. Bock */
	class vinfo
	{
	public:
		/// virtual table offset
		size_t	t_offs;
		/// distance to base class
		int		dist;

		/** \param to table offset
			\param d distance to base class */
		vinfo( size_t to, int d );
	};

	typedef vector< Table* > tbl_vec;
	typedef map< Class*, vinfo > vinfo_map;
	typedef set< Class* > class_set;

	/// resolved class name
	string		name;
	/// size
	size_t		size;
	/// 'builtin type' flag
	bool		bit;
	/// 'tainted class' flag
	bool		tainted;
	/// virtual table
	VTable*		vtbl;
	/// frost tables
	tbl_vec		tables;
	/// derived classes
	class_set	derivs;
	/// virtual table / base info map
	vinfo_map	bases;
	/// direct and indirect virtual base classes
	class_set	vbases;

	/** destructor */
	~Class();
	
	/// forbidden
	Class( const Class& );
	
	/// forbidden
	void
	operator = ( const Class& );

	friend void
	avoid_stupid_gcc_warning();
};

#endif // _frost_class_hh
