/*
 * frost - C++ multi methods as if they were a native feature
 * Copyright (C) 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */

#ifndef _frost_function_hh
#define _frost_function_hh


// includes

#include <uds/refcounter.hh>
#include <string>
#include <vector>
#include "type.hh"

using namespace std;
using namespace uds;


/** function class.
	\author Julian v. Bock */
class Function : public RefCounter<>
{
public:
	typedef GC_Ptr< Function > ptr;
	typedef vector< Type > type_vec;
	typedef vector< Class* > cl_vec;
	typedef vector< size_t > so_vec;
	
	/** \param en encoded function name */
	Function( const string& en );

	/** \return encoded function name */
	const string&
	EName() const { return ename; }

	/** returns the complete, demangled function name. 'static' and 'virtual'
		keywords are inserted where necessary.
		\param rm if set to true, the additional information added to the
		function name by frost is removed
		\return demangled function name */
	string
	DemName( bool rm = false ) const;

	/** \return function name (not encoded) */
	const string&
	FName() const { return fname; }

	/** sets the function name (not encoded)
		\param fn function name */
	void
	SetFName( const string& fn );

	/** adds a function argument
		\param tp type information */
	void
	AddArg( const Type& tp );

	/** \return arguments */
	const type_vec&
	Args() const { return args; }

	/** \return number of arguments (including implicit this or dummy
		argument) */
	size_t
	ArgCount() const { return args.size(); }

	/** \return virtual arg class vector */
	cl_vec
	VAClassVec() const;

	/** \return virtual stack offset vector */
	so_vec
	VASOffsVec() const;

	/** \return true if the function is a multi method */
	bool
	IsMM() const { return ismm; }

	/** \return true if the function is a static member function */
	bool
	IsStatic() const;

	/** \return true if the function is const */
	bool
	IsConst() const { return isc; }

	/** sets the function to const */
	void
	SetConst() { isc = true; }

	/** \return true if the function pure virtual */
	bool
	IsPure() const { return isp; }

	/** sets the pure virtual function flag */
	void
	SetPure() { isp = true; }

	/** checks whether the type is correct for all virtual arguments */
	void
	SanityCheck() const;

	/** compares the function arguments with the arguments of another
		function. this method is used to determine whether this function
		is higher in the hierarchy of a multi method group.
		\param fn function to compare
		\return -1 if the functions are not compatible, 1 if this function
		is higher in the hierarchy of a multi method group. Otherwise
		0 is returned. */
	int
	MMG_Compare( const Function& fn ) const;

	/** output stream operator
		\param out output stream
		\param fn function */
	friend ostream&
	operator << ( ostream& out, const Function& tp );

	bool
	operator < ( const Function& rhs ) const { return ename < rhs.ename; }

private:
	/// function name
	string		fname;
	/// encoded function name
	string		ename;
	/// function arguments
	type_vec	args;
	/// 'is multi method' flag
	bool		ismm;
	/// 'is const method' flag
	bool		isc;
	/// 'is pure virtual' flag
	bool		isp;

	/** destructor */
	virtual
	~Function();

	friend void
	avoid_stupid_gcc_warning();
};

#endif // _frost_function_hh
