/*
 * Frost - use C++ multi methods as if they were a native feature
 * Copyright (C) 1999, 2000, 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <uds/filestream.hh>
#include <uds/functional.hh>
#include <uds/io_util.hh>
#include <uds/procstream.hh>
#include <uds/stl_util.hh>
#include <uds/str_util.hh>
#include <uds/sys_util.hh>
#include <bfd.h>
#include <memory>
#include "ascode.hh"
#include "class.hh"
#include "dbase.hh"
#include "exception.hh"
#include "fnfilt.hh"
#include "hierarchy.hh"
#include "mm.hh"
#include "progb.hh"
#include "sio.hh"
#include "table.hh"
#include "titable.hh"
#include "util.hh"
#include "vtable.hh"

using namespace uds;


namespace
{

class cl_info
{
public:
	const char* name;
	size_t size;
};


void
print_code_header( ostream& out )
{
	out<<"# This file was generated by Frost version "
	   <<VERSION<<" ("<<VERSION_TAG<<")\n"
		"# \n"
		"# IT IS PROVIDED \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED\n"
		"# OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n"
		"# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.\n"
		"\n";
}

inline void
insert_func_maybe( set< string >& funcs, const string& sym )
{
	if ( sym.substr( 0, 2 ) == "_Z" )
		funcs.insert( sym );
}

void
throw_bfd_error()
{
	throw runtime_error( bfd_errmsg( bfd_get_error() ));
}

}


// Hierarchy implementation

Hierarchy::Hierarchy()
{
	cl_info bi_types[] = {
		{ "bool",				sizeof( bool ) },
		{ "char",				sizeof( char ) },
		{ "signed char",		sizeof( signed char ) },
		{ "unsigned char",		sizeof( unsigned char ) },
		{ "wchar_t",			sizeof( wchar_t ) },
		{ "short",				sizeof( short ) },
		{ "unsigned short",		sizeof( unsigned short ) },
		{ "int",				sizeof( int ) },
		{ "unsigned",			sizeof( unsigned ) },
		{ "long",				sizeof( long ) },
		{ "unsigned long",		sizeof( unsigned long ) },
		{ "long long",			sizeof( long long ) },
		{ "unsigned long long",	sizeof( unsigned long long ) },
		{ "float",				sizeof( float ) },
		{ "double",				sizeof( double ) },
		{ "long double",		sizeof( long double ) },
		{ 0, } };

	for ( cl_info* ci = bi_types; ci->name; ++ci )
		MapInsert< internal_error >( classes, ci->name,
									 new Class( ci->name, ci->size, true ) );
}

Hierarchy::~Hierarchy()
{
	for ( class_map::iterator i = classes.begin(); i != classes.end(); ++i )
		i->second->Destroy();

	for ( vtbl_map::iterator i = vtbls.begin(); i != vtbls.end(); ++i )
		delete i->second;
}

Class&
Hierarchy::AddClass( const string& name )
{
	Class* cl = new Class( name, 0, false );
	MapInsert< internal_error >( classes, name, cl );
	dcl.push_back( cl );

	return *cl;
}

Class&
Hierarchy::GetClass( const string& n )
{
	return *MapFind< internal_error >( classes, n );
}

Class*
Hierarchy::QueryClass( const string& n )
{
	return SafeDeref( MapQuery( classes, n ) );
}

void
Hierarchy::AddFunc( Function::ptr fn )
{
	string en = fn->EName();

	if ( !fn->IsMM() )
	{
		_frost_Log( LOG_DEBUG + 1, "dropping function " + en
					+ " (not a fwva group member)" );
	}
	else if ( fnames.insert( en ).second )
	{
		_frost_Log( LOG_DEBUG + 1, "adding function " + en );
		funcs.insert( fn );
	}
	else
	{
		_frost_Log( LOG_DEBUG + 1, "dropping function " + en
					+ " (already defined)" );
	}
}

VTable&
Hierarchy::AddVTable( const string& n, Class& o, size_t s )
{
	VTable* v = new VTable( n, o, s );
	MapInsert< internal_error >( vtbls, n, v );

	return *v;
}

bool
Hierarchy::HasMM() const
{
	for ( func_set::const_iterator i = funcs.begin(); i != funcs.end(); ++i )
		if ( ( *i )->IsMM() )
			return true;

	return false;
}

bool
Hierarchy::DumpMMs( ostream& out )
{
	typedef multimap< string, MMethod* > mm_mmap_t;
	typedef mm_mmap_t::iterator mmi;
	typedef map< MMethod::m_ptr, MMethod* > mm_map_t;
	typedef mm_map_t::iterator mi;
	typedef vector< DispBase* > disp_vec;
	
	mm_mmap_t	mm_mmap;
	mm_map_t	mm_map;
	disp_vec	disps;

	// stage 1: construct multi method groups

	for ( func_set::const_iterator i = funcs.begin(); i != funcs.end(); ++i )
	{
		Function& fn = **i;
		if ( !fn.IsMM() )
			continue;

		fn.SanityCheck();

		pair< mmi, mmi > range = mm_mmap.equal_range( fn.FName() );
		bool mfound = false;
		
		for ( mmi j = range.first; j != range.second; ++j )
		{
			MMethod& mm = *j->second;
			bool m = mm.AddMemberMaybe( fn );

			if ( m )
				mfound = true;
		}

		if ( !mfound )
		{
			// this function does not belong to an existing multi method group.
			// create a new one
			
			mm_mmap.insert( make_pair( fn.FName(), new MMethod( fn, tgrp ) ));
		}
	}

	// stage 2: all multi method groups are created. collect mm members a
	// second time.

	for ( func_set::const_iterator i = funcs.begin(); i != funcs.end(); ++i )
	{
		Function& fn = **i;
		if ( !fn.IsMM() )
			continue;

		pair< mmi, mmi > range = mm_mmap.equal_range( fn.FName() );
		for ( mmi j = range.first; j != range.second; ++j )
		{
			MMethod& mm = *j->second;
			mm.AddMemberMaybe( fn );
		}
	}

	// stage 3: collect fwvas in a map to remove duplicates

	for ( mmi i = mm_mmap.begin(); i != mm_mmap.end(); ++i )
		mm_map[ i->second->Root() ] = i->second;

	// stage 4: process all multi method groups

	bool succ = true;
	for ( mi i = mm_map.begin(); i != mm_map.end(); ++i )
	{
		_frost_Log( LOG_DEBUG + 1, "processing mm group "
					+ i->second->Name() );
		
		succ &= i->second->BuildMM( disps );
	}

	// stage 5: arrange frost table fragments

	tgrp.ArrangeFrags();

	// stage 6: dump assembly code

	print_code_header( out );

	set< Class* > vcs;
	tgrp.FWVAClasses( vcs );
	for ( set< Class* >::const_iterator i = vcs.begin(); i != vcs.end(); ++i )
	{
		Class& cl = **i;
		
		if ( cl.Tainted() )
			throw error( "unable to build FWVAs for tainted class "
						 + cl.Name() + "\n"
						 "virtual table information is not available" );
		
		cl.Vtbl().EmitCode( out );
	}
	
	out<<endl;
	tgrp.EmitCode( out );
	EnterTextSegment( out );

	// print the noreturn function dummy code

	EmitNoreturn( out );

	// print the dispatcher assembly code. duplicate dispatchers are dropped.
	// this happens when one fwva has more than one root because of
	// multiple inheritance.
	// this can be optimized to use the dispatcher which contains less
	// overhead.

	{
		set< string > ds;
		
		for ( size_t i = 0; i < disps.size(); ++i )
		{
			DispBase& d = *disps[i];
			string dn = d.Name();
			
			if ( d.Global() && !ds.insert( dn ).second )
				_frost_Log( LOG_DEBUG + 1,
							"dropping duplicate dispatcher " + dn );
			else
				d.EmitCode( out );
		}
	}

	// cleanup

	for ( mmi i = mm_mmap.begin(); i != mm_mmap.end(); ++i )
		delete i->second;

	for ( size_t i = 0; i < disps.size(); ++i )
		delete disps[i];

	return succ;
}

void
Hierarchy::Read( istream& in )
{
	FinalAction ex_restore( function( &istream::exceptions, value( &in ),
									  value( in.exceptions() )));

	in.exceptions( ios::badbit | ios::failbit | ios::eofbit );

	try
	{
		const char* magic = "(frost-object ";
		const size_t mg_sz = strlen( magic );
		
		char magic_test[mg_sz];
		in.read( magic_test, mg_sz );

		if ( string( magic_test, mg_sz ) != magic )
			throw InvFOFile( "bad magic string" );

		// read classes

		WalkList( in, function( &Hierarchy::read_class, value( this ),
								reference( in ) ), "classes" );

		// read frost table hierarchies

		WalkList( in, function( &Hierarchy::read_fthier, value( this ),
								reference( in ) ), "ft-hier" );

		// read virtual tables

		WalkList( in, function( &Hierarchy::read_vtable, value( this ),
								reference( in ) ), "vtables" );

		// read functions

		WalkList( in, function( &Hierarchy::read_function, value( this ),
								reference( in ) ), "functions" );

		if ( in.get() != ')' )
			in.clear( ios::failbit );
	}
	catch ( ios::failure& )
	{
		throw InvFOFile( "file corrupted" );
	}

	while ( true )
	{
		char c = in.peek();

		if ( c == EOF )
			break;

		if ( c != ' ' && c != '\t' && c != '\n' )
			throw InvFOFile( "file corrupted - EOF expected" );

		in.get();
	}
}

void
Hierarchy::ReadELF( const string& f )
{
	ProcStream ps( "file -bL " + ShellQuote( f ) );
	string ft = ReadLine( ps );

	_frost_Log( LOG_DEBUG + 1, f + ": " + ft );

	if ( ft.find( "relocatable" ) != string::npos )
		read_elf_o( f );
	else if ( ft.find( "ar archive" ) != string::npos )
		read_elf_a( f );
	else if ( ft.find( "shared object" ) != string::npos )
		read_elf_so( f );
	else
		throw UserError( f + ": " + ft + "\nunknown format" );
}

void
Hierarchy::process_flist( const string& f, fn_set& fns )
{
	// build function list
	
	string flist = ( Self().Flags( ProgState::keep_fl ) )
		? Self().RepSuffix( f, ".flist", true ) : Self().TmpFileName();
	FileStream fls( flist, ios::in | ios::out | ios::trunc );

	for ( fn_set::const_iterator i = fns.begin(); i != fns.end(); ++i )
	{
		string ds = DemSym( *i );
		if ( ds.find( '(' ) != string::npos )
			fls<<*i<<' '<<ds<<endl;
	}

	// process function list

	_frost_Log( LOG_DEBUG + 1, "processing function list " + flist );

	FnFilt fnfilt( &fls, *this );
	fnfilt.yylex();
}

void
Hierarchy::read_class( istream& in )
{
	ListMarker lm( in );

	string name = ReadQStr( in );
	size_t size;
	bool tainted;
	
	in>>size;
	tainted = ReadBool( in );

	if ( classes.find( name ) != classes.end() )
	{
		_frost_Log( LOG_DEBUG, "skipping class " + name );

		SkipList( in );
		in.unget();
		return;
	}

	Class& cl = *new Class( name, size, false );
	if ( tainted )
		cl.Taint();

	WalkList( in, function( &Hierarchy::read_ftable, value( this ),
							reference( in ), reference( cl ) ));

	WalkList( in, function( &Hierarchy::read_base, value( this ),
							reference( in ), reference( cl ) ));

 	WalkList( in, function( &Hierarchy::read_vbase, value( this ),
							reference( in ), reference( cl ) ));

	MapInsert< internal_error >( classes, name, &cl );
	dcl.push_back( &cl );
}

void
Hierarchy::read_ftable( istream& in, Class& cl )
{
	ListMarker lm( in );

	int fv_offs;
	string name = ReadQStr( in );
	in>>fv_offs;

	Table& tbl = cl.AddFTable( name, tgrp, fv_offs );
	MapInsert< internal_error >( ftbls, name, &tbl );
}

void
Hierarchy::read_base( istream& in, Class& cl )
{
	ListMarker lm( in );

	string tbl = ReadQStr( in );
	Class* base = ( tbl == cl.Name() ) ? &cl : &GetClass( tbl );

	size_t	t_offs;
	int		dist;
				
	in>>t_offs>>dist;

	cl.SetTableOffset( *base, t_offs, dist );
}

void
Hierarchy::read_vbase( istream& in, Class& cl )
{
	string vb = ReadQStr( in );

	cl.AddVBase( GetClass( vb ) );
}

void
Hierarchy::read_fthier( istream& in )
{
	string name = ReadQStr( in );
	Table& ptbl = *MapFind< internal_error >( ftbls, name );

	WalkList( in, function( &Hierarchy::read_fthier_p, value( this ),
							reference( in ), reference( ptbl ) ));
}

void
Hierarchy::read_fthier_p( istream& in, Table& p )
{
	string name = ReadQStr( in );

	Table& tbl = *MapFind< internal_error >( ftbls, name );
	tbl.SetParent( p );

	WalkList( in, function( &Hierarchy::read_fthier_p, value( this ),
							reference( in ), reference( tbl ) ));
}

void
Hierarchy::read_vtable( istream& in )
{
	ListMarker lm( in );

	string owner = ReadQStr( in );
	VTable* vtbl = 0;

	{
		ListMarker lm( in );

		string name = ReadQStr( in );
		size_t sz;
		in>>sz;

		vtbl_map::iterator i = vtbls.find( name );
		if ( i != vtbls.end() && i->second->Complete() )
		{
			_frost_Log( LOG_DEBUG, "skipping vtable " + name );

			SkipList( in );
			if ( in.peek() != ')' )
			{
				while ( in.get() != '(' );
				SkipList( in );
			}
			in.unget();
			
			return;
		}

		vtbl = new VTable( name, GetClass( owner ), sz );

		WalkList( in, function( &Hierarchy::read_vti_delta, value( this ),
								reference( in ), reference( *vtbl ) ));

		WalkList( in, function( &Hierarchy::read_vti_slot, value( this ),
								reference( in ), reference( *vtbl ) ));

		vtbl->SetComplete();

		if ( i != vtbls.end() )
		{
			delete i->second;
			i->second = vtbl;
		}
		else
			vtbls[name] = vtbl;
	}
	
	if ( in.peek() != ')' )
	{
		ListMarker lm( in );

		string name = ReadQStr( in );
		size_t sz;
		in>>sz;
		
		TITable& tit = *new TITable( name, *vtbl );

		WalkList( in, function( &Hierarchy::read_vti_delta, value( this ),
								reference( in ), reference( tit ) ));
		
		WalkList( in, function( &Hierarchy::read_vti_slot, value( this ),
								reference( in ), reference( tit ) ));

		tit.SetComplete();
	}
}

void
Hierarchy::read_vti_slot( istream& in, vti_base& tbl )
{
	ListMarker lm( in );

	int s;
	in>>s;
	string v = ReadQStr( in );

	tbl.Assign( s, v );
}

void
Hierarchy::read_vti_delta( istream& in, vti_base& tbl )
{
	int d;
	in>>d;

	tbl.Fill( d );
}

void
Hierarchy::read_function( istream& in )
{
	ListMarker lm( in );

	string name = ReadQStr( in );
	string ename = ReadQStr( in );
	bool cm = ReadBool( in );
	bool pm = ReadBool( in );

	Function::ptr fn = new Function( ename );
	fn->SetFName( name );
	if ( cm )
		fn->SetConst();
	if ( pm )
		fn->SetPure();

	WalkList( in, function( &Hierarchy::read_fn_arg, value( this ),
							reference( in ), value( fn ) ));

	AddFunc( fn );
}

void
Hierarchy::read_fn_arg( istream& in, Function::ptr fn )
{
	ListMarker lm( in );

	string	cn = ReadQStr( in );
	string	sr = ReadQStr( in );
	int		sz;
	in>>sz;
	bool	va = ReadBool( in );

	fn->AddArg( Type( *this, cn, sr, sz, va ) );
}

void
Hierarchy::read_elf_o( const string& f )
{
	_frost_Log( LOG_NOTICE, "processing object file " + f );

	// open object file

	bfd* file = bfd_openr( f.c_str(), 0 );
	if ( !file )
		throw_bfd_error();

	if ( !bfd_check_format( file, bfd_object ) )
		throw_bfd_error();

	// fetch symtab

	_frost_Log( LOG_INFO, "reading symbol table" );

	if ( !( bfd_get_file_flags( file ) & HAS_SYMS ) )
		return;

	long st_size = bfd_get_symtab_upper_bound( file );
	if ( st_size < 0 )
		throw_bfd_error();

	if ( !st_size )
		return;

	auto_ptr< asymbol* > symtab( static_cast< asymbol** >(
									 operator new ( st_size ) ));
	long sym_cnt = bfd_canonicalize_symtab( file, symtab.get() );

	if ( !sym_cnt )
		return;

	// process symtab

	_frost_Log( LOG_INFO, "processing symbol table" );

	fn_set fns;
	
	for ( long i = 0; i < sym_cnt; ++i )
	{
		const char* head = ".gnu.linkonce.t.";
		const size_t heads = strlen( head );

		if ( !symtab.get()[i] )
			continue;

		string sym = symtab.get()[i]->name;
		if ( sym.empty() )
			continue;

		string sec = symtab.get()[i]->section->name;
		
		// get functions from symbol table

		if ( sec == ".text" || sec.substr( 0, heads ) == head )
			insert_func_maybe( fns, sym );
	}

	// process sections

	for ( asection* section = file->sections; section;
		  section = section->next )
	{
		if ( !( section->flags & SEC_HAS_CONTENTS ) )
			continue;
		
		string sec = section->name;
		const char* vthead = ".gnu.linkonce.d._ZTV";
		const char* tihead = ".gnu.linkonce.d._ZTI";
		const size_t hlen = strlen( tihead );
		string tname;
		vti_base* tbl;

		// check whether the section needs to be scanned

		if ( sec.substr( 0, hlen ) == tihead )
		{
			// typeinfo table section
			
			tname = sec.substr( hlen - 4 );
			string vtname = tname;
			vtname.at( 3 ) = 'V';
			
			VTable* vtbl = SafeDeref( MapQuery( vtbls, vtname ) );
			if ( !vtbl )
			{
				_frost_Log( LOG_DEBUG + 1, "skipping typeinfo table " +  tname
							+ " for unknown virtual table " + vtname );
				continue;
			}
			else if ( vtbl->HasTInfo() )
			{
				_frost_Log( LOG_DEBUG + 1, "skipping typeinfo table " +  tname
							+ "; virtual table " + vtname + " has already "
							"typeinfo table" );
				continue;
			}

			tbl = new TITable( tname, *vtbl );
		}
		else if ( sec.substr( 0, hlen ) == vthead )
		{
			// virtual table section
			
			tname = sec.substr( hlen - 4 );
			tbl = SafeDeref( MapQuery( vtbls, tname ) );

			if ( !tbl )
			{
				_frost_Log( LOG_DEBUG + 1, "skipping unknown virtual table "
							+ tname );
				continue;
			}
			else if ( tbl->Complete() )
			{
				_frost_Log( LOG_DEBUG + 1, "skipping complete virtual table "
							+ tname );
				continue;
			}
		}
		else
		{
			// nothing interesting
			
			continue;
		}

		// process raw section contents

		_frost_Log( LOG_DEBUG, "filling table " + tname + " (raw contents)" );

		long sz = bfd_section_size( file, section );
		if ( !sz )
			continue;
		
		auto_ptr< int > cont( static_cast< int* >( operator new ( sz ) ));
		if ( !bfd_get_section_contents( file, section, cont.get(), 0, sz ) )
			throw_bfd_error();

		for ( long i = 0, j = 0; j < sz; ++i, j += sizeof( void* ) )
			tbl->Fill( cont.get()[i] );

		// relocations

		_frost_Log( LOG_DEBUG, "processing relocations" );

		sz = bfd_get_reloc_upper_bound( file, section );
		if ( sz < 0 )
			throw_bfd_error();

		auto_ptr< arelent* > rels(
			static_cast< arelent** >( operator new ( sz ) ));
		long relc = bfd_canonicalize_reloc( file, section, rels.get(),
											symtab.get() );

		if ( relc < 0 )
			throw_bfd_error();
		else if ( relc == 0 )
			continue;

		for ( long i = 0; i < relc; ++i )
		{
			const char* sym = ( *rels.get()[i]->sym_ptr_ptr )->name;
			size_t addr = rels.get()[i]->address;

			if ( addr % sizeof( void* ) )
				_uds_abort;

			tbl->Assign( addr / sizeof( void* ), sym );
		}

		// mark vtable as complete

		_frost_Log( LOG_DEBUG, "table complete" );

		tbl->SetComplete();
	}

	bfd_close( file );
	process_flist( f, fns );
}

void
Hierarchy::read_elf_a( const string& f )
{
	_frost_Log( LOG_NOTICE, "processing static library " + f );

	ProcStream flist( "ar t " + ShellQuote( f ) );
	fn_set fns;

	while ( true )
	{
		string file;
		flist>>file;

		if ( !flist )
			break;

		string obj = Self().TmpFileName();
		string cmd = "ar p " + ShellQuote( f ) + " " + ShellQuote( file )
			+ " > " + obj;

		_frost_Log( LOG_DEBUG, "executing " + cmd );
		
		System( cmd.c_str() );
		read_elf_o( obj );
	}

	process_flist( f, fns );
}

void
Hierarchy::read_elf_so( const string& f )
{
	_frost_Log( LOG_NOTICE, "processing shared object " + f );

	// open shared object file

	bfd* file = bfd_openr( f.c_str(), 0 );
	if ( !file )
		throw_bfd_error();

	if ( !bfd_check_format( file, bfd_object ) )
		throw_bfd_error();

	// fetch symtab

	_frost_Log( LOG_INFO, "reading symbol table" );

	if ( !( bfd_get_file_flags( file ) & HAS_SYMS ) )
		return;

	long st_size = bfd_get_dynamic_symtab_upper_bound( file );
	if ( st_size < 0 )
		throw_bfd_error();

	if ( !st_size )
		return;

	auto_ptr< asymbol* > symtab( static_cast< asymbol** >(
									 operator new ( st_size ) ));
	long sym_cnt = bfd_canonicalize_dynamic_symtab( file, symtab.get() );

	if ( !sym_cnt )
		return;

	// process symtab

	_frost_Log( LOG_INFO, "processing symbol table" );

	typedef map< size_t, pair< vti_base*, size_t > > vt_map_t;

	fn_set fns;
	vt_map_t vt_map;
	
	for ( long i = 0; i < sym_cnt; ++i )
	{
		asymbol* symbol = symtab.get()[i];

		if ( !symbol )
			continue;

		string sym = symbol->name;
		if ( sym.empty() )
			continue;
		
		// get functions and virtual / typeinfo tables from symbol table

		string sub = sym.substr( 0, 4 );
		vti_base* tbl;
		
		if ( sub == "_ZTI" )
		{
			// typeinfo table
			
			string vtname = sym;
			vtname.at( 3 ) = 'V';
			
			VTable* vtbl = SafeDeref( MapQuery( vtbls, vtname ) );
			if ( !vtbl )
			{
				_frost_Log( LOG_WARNING, "skipping typeinfo table " +  sym
							+ " for unknown virtual table " + vtname );
				continue;
			}
			else if ( vtbl->HasTInfo() )
			{
				_frost_Log( LOG_DEBUG + 1, "skipping typeinfo table " +  sym
							+ "; virtual table " + vtname + " has already "
							"typeinfo table" );
				continue;
			}

			tbl = new TITable( sym, *vtbl );
		}
		else if ( sub == "_ZTV" )
		{
			// virtual table

			tbl = SafeDeref( MapQuery( vtbls, sym ) );

			if ( !tbl )
			{
				_frost_Log( LOG_DEBUG + 1, "skipping unknown virtual table "
							+ sym );
				continue;
			}
			else if ( tbl->Complete() )
			{
				_frost_Log( LOG_DEBUG + 1, "skipping complete virtual table "
							+ sym );
				continue;
			}
		}
		else
		{
			insert_func_maybe( fns, sym );
			continue;
		}

		symbol_info si;
		bfd_symbol_info( symbol, &si );
		size_t addr = si.value;
		size_t size = symbol->value;
			
		_frost_Log( LOG_DEBUG + 2, "table address " + StringVar( addr )
					+ "; size " + StringVar( size ) );
			
		vt_map[addr] = make_pair( tbl, size );
	}

	if ( !vt_map.empty() ) do
	{
		// get the .data section

		asection* data_section = 0;
		size_t offset = 0;
	
		for ( asection* section = file->sections; section;
			  section = section->next )
		{
			if ( section->name == ".data" )
			{
				_frost_Log( LOG_DEBUG, "located .data section at offset "
							+ StringVar( offset ) );
			
				data_section = section;
				break;
			}

			offset += bfd_section_size( file, section );
		}

		if ( !data_section || !( data_section->flags & SEC_HAS_CONTENTS ) )
			break;

		// get .data section data

		long sz = bfd_section_size( file, data_section );
		if ( !sz )
			continue;
		
		auto_ptr< int > cont( static_cast< int* >( operator new ( sz ) ));
		if ( !bfd_get_section_contents( file, data_section, cont.get(), 0,
										sz ) )
			throw_bfd_error();

		if ( vt_map.rbegin()->first + vt_map.rbegin()->second.second
				 >= offset + sz )
				_uds_abort;

		// get dynamic relocations

		_frost_Log( LOG_DEBUG, "processing relocations" );

		sz = bfd_get_dynamic_reloc_upper_bound( file );
		if ( sz < 0 )
			throw_bfd_error();

		auto_ptr< arelent* > rels(
			static_cast< arelent** >( operator new ( sz ) ));
		long relc = bfd_canonicalize_dynamic_reloc( file, rels.get(),
													symtab.get() );

		if ( relc < 0 )
			throw_bfd_error();
		else if ( relc == 0 )
			break;

		// process all tables

		long cur_rel = 0;
		for ( vt_map_t::iterator i = vt_map.begin(); i != vt_map.end(); ++i )
		{
			size_t addr = i->first;
			vti_base* tbl = i->second.first;
			size_t size = i->second.second;

			_frost_Log( LOG_DEBUG, "filling table " + tbl->Name()
						+ " (raw contents)" );

			for ( long i = 0, j = 0; j < sz; ++i, j += sizeof( void* ) )
				tbl->Fill( cont.get()[i] );

			_frost_Log( LOG_DEBUG, "processing relocations" );

			while ( true )
			{
				if ( cur_rel == relc )
				{
					// no more relocations
					break;
				}

				const char* rsym = ( *rels.get()[cur_rel]->sym_ptr_ptr )->name;
				size_t raddr = rels.get()[cur_rel]->address;

				if ( raddr % sizeof( void* ) )
					_uds_abort;

				if ( raddr > addr + size )
				{
					// no more relocations for this table
					break;
				}

				++cur_rel;

				if ( raddr < addr )
				{
					// relocation does not belong to this table
					continue;
				}

				tbl->Assign( raddr / sizeof( void* ), rsym );
			}

			// mark vtable as complete

			_frost_Log( LOG_DEBUG, "table complete" );

			tbl->SetComplete();
		}
	} while ( false );

	bfd_close( file );
	process_flist( f, fns );
}


// misc. implementation

ostream&
operator << ( ostream& out, const Hierarchy& h )
{
	out<<"(frost-object (classes";

	for ( size_t i = 0; i < h.dcl.size(); ++i )
	{
		out<<' '<<*h.dcl[i];
	}

	out<<") ";
	h.tgrp.DumpHierarchies( out );
	out<<" (vtables";

	for ( Hierarchy::vtbl_map::const_iterator i = h.vtbls.begin();
		  i != h.vtbls.end(); ++i )
		out<<' '<<*i->second;
	
	out<<") (functions";

	for ( Hierarchy::func_set::const_iterator i = h.funcs.begin();
		  i != h.funcs.end(); ++i )
	{
		out<<' '<<**i;
	}

	return out<<"))";
}
