/*
 * frost - C++ multi methods as if they were a native feature
 * Copyright (C) 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <uds/procstream.hh>
#include <uds/str_util.hh>
#include <uds/sys_util.hh>
#include <fcntl.h>
#include <fstream>
#include <unistd.h>
#include "clfilt.hh"
#include "exception.hh"
#include "fpp.hh"
#include "hierarchy.hh"
#include "prog.hh"


// ProgramState implementation

void
ProgramState::ProcInputFiles()
{
	log( LOG_NOTICE, "processing input files" );

	for ( cstage_t ss = 0; ss <= lcs; ++ss )
	{
		fileq_t& pfq = ifq.at( ss );
		
		while ( pfq.size() )
		{
			string file = pfq.front();
			pfq.pop();

			proc_file( file, ss );
		}
	}

	if ( lcs == link )
		link_files();
}

void
ProgramState::read_fobj( Hierarchy& hier, const string& file )
{
	ifstream in( file.c_str(), ios::in );
	if ( !in )
		throw system_error( file );
		
	_frost_Log( LOG_INFO, "reading frost object file " + file );

	try
	{
		hier.Read( in );
	}
	catch ( InvFOFile& e )
	{
		e.Info() = file + ": " + e.Info();
		throw;
	}
}

void
ProgramState::proc_file( const string& file, cstage_t ss )
{
	log( LOG_NOTICE, "processing file " + file );

	// preprocessed file
	string pp_file = file;
	// compiled file
	string cp_file = file;
	// assembled file
	string as_file = file;
	// frost object file
	string fo_file;
	// hierarchy object
	Hierarchy hier;

	switch ( ss )
	{
	case preprocess:
	{
		log( LOG_INFO, "entering preprocessing stage" );

		string cmd = gcc_cmd + cpp_opts + "-E " + ShellQuote( file );
		log( LOG_INFO, "launching FrostPP; using '" + cmd + "' as input" );

		// determine output stream. If preprocessing is the last compile
		// stage, dump output to stdout

		FileStream fs;
		ostream* out = &cout;
		if ( lcs > preprocess || !out_file.empty() )
		{
			pp_file = cstage_output_filename( file, preprocess );
			fs.Open( pp_file, ios::in | ios::out | ios::trunc );
			out = &fs;
		}

		ProcStream in( cmd );
		FrostPP fpp( &in, out, file );
		fpp.yylex();
	}
	case compile:
	{
		if ( lcs < compile )
			break;

		log( LOG_INFO, "entering compilation stage" );

		// add the class file to the list of temporary files

		string cl_file = pp_file.substr( pp_file.rfind( "/" ) + 1,
										 string::npos ) + ".class";

		if ( ~flags & keep_cl )
			tfiles.push_back( cl_file );

		// compile source code and generate class file

		cp_file = cstage_output_filename( file, compile );
		exec_gcc( "-x c++-cpp-output -S -fdump-class-hierarchy "
				  "-fkeep-inline-functions " + ShellQuote( pp_file ) + " -o "
				  + cp_file );

		// parse class file

		try
		{
			log( LOG_DEBUG, "parsing class file" );
			{
				FileStream in( cl_file, ios::in );
				ClFilt filt( &in, hier );
				filt.yylex();
			}
		}
		catch ( system_error& e )
		{
			log( LOG_WARNING, "unable to open file " + cl_file );
		}

		// dump hierarchy to frost object file

		fo_file = fo_filename( file );

		if ( lcs < assemble )
		{
			FileStream out( fo_file, ios::out | ios::trunc );
			out<<hier;
		}
	}
	case assemble:
	{
		if ( lcs < assemble )
			break;

		log( LOG_INFO, "entering assembly stage" );

		if ( fo_file.empty() )
		{
			fo_file = fo_filename( file );
			
			if ( FileExist( fo_file.c_str() ))
				 read_fobj( hier, fo_file );
		}

		as_file = cstage_output_filename( file, assemble );
		exec_gcc( "-x assembler -c " + ShellQuote( cp_file ) + " -o "
				  + ShellQuote( as_file ), true );

		try
		{
			hier.ReadELF( as_file );
			FileStream out( fo_file, ios::out | ios::trunc );
			out<<hier;
		}
		catch ( ... )
		{
			// remove the compiler output file in case of a failure
			// to make sure that it will be rebuilt by make the next time

			unlink( as_file.c_str() );
			throw;
		}
	}
	case link:
	{
		if ( lcs < link )
			break;

		log( LOG_INFO, "entering link stage" );

		if ( fo_file.empty() )
		{
			// check whether there is a frost object file
			
			fo_file = RepSuffix( file, ".fo", false );
			int fd = open( fo_file.c_str(), O_RDONLY );

			if ( fd >= 0 )
				close( fd );
			else
				fo_file.erase();
		}
		
		linkq.push_back( as_file );
	}
	}
	
	if ( fo_file.empty() )
		relfq.push_back( as_file );
	else
		fofq.push_back( fo_file );
}

void
ProgramState::link_files()
{
	log( LOG_INFO, "linking files" );

	// master hierarchy
	Hierarchy hier;

	// log library paths
	
	if ( ::Log( LOG_INFO ) )
	{
		::Log( LOG_INFO, "library paths:" );
			
		for ( size_t i = 0; i < libpaths.size(); ++i )
			::Log( LOG_INFO, libpaths[i] );
	}

	if ( linkq.empty() )
		_uds_abort;
	else if ( linkq.size() == 1 )
		throw UserError( "no linker input files" );

	// read all frost object files

	for ( size_t i = 0; i < fofq.size(); ++i )
		read_fobj( hier, fofq[i] );

	// read object files and libraries to complete virtual tables

	for ( size_t i = 0; i < relfq.size(); ++i )
		hier.ReadELF( relfq[i] );

	// dump final frost object file if requested

	if ( flags & dump_ffo )
	{
		FileStream out( "frost.out.fo", ios::out | ios::trunc );
		out<<hier;
	}
	
	// build multi methods

	_frost_Log( LOG_INFO, "building multi methods" );

	string as_file = ( flags & keep_asm ) ? "frost.out.s" : TmpFileName();
	FileStream out( as_file, ios::out | ios::trunc );
		
	hier.DumpMMs( out );
	out.Close();

	string fo_file = Flags( keep_asm | save_tmp )
		? "frost.out.o" : TmpFileName();
	exec_gcc( "-x assembler -c " + ShellQuote( as_file ) + " -o "
			  + ShellQuote( fo_file ) );

	// build gcc command line
	
	string cmd = "-o " + ShellQuote( cstage_output_filename( "", link ) ) + " "
		+ ShellQuote( fo_file );
	for ( size_t i = 0; i < linkq.size(); ++i )
		cmd += " " + ShellQuote( linkq[i] );

	exec_gcc( cmd );
}
